
# This file was generated by 'versioneer.py' (0.26) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-03-24T13:53:30-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "505ea813fbc9d71a80faf63b1f83928fcfa77e97",
 "version": "23.02.01"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
