# Copyright (c) 2023, NVIDIA CORPORATION.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from .binding.pylibcugraphops_ext import (
    fg_csr_int32,
    fg_csr_int64,
    mfg_csr_int32,
    mfg_csr_int64,
    fg_csr_hg_int32,
    fg_csr_hg_int64,
    mfg_csr_hg_int32,
    mfg_csr_hg_int64,
)

FgCsr = Union[fg_csr_int32, fg_csr_int64]
MfgCsr = Union[mfg_csr_int32, mfg_csr_int64]
AnyCsr = Union[FgCsr, MfgCsr]
CuGraphFgCsr = FgCsr
CuGraphMfgCsr = MfgCsr
CuGraphAnyCsr = AnyCsr

FgHgCsr = Union[fg_csr_hg_int32, fg_csr_hg_int64]
MfgHgCsr = Union[mfg_csr_hg_int32, mfg_csr_hg_int64]
AnyHgCsr = Union[FgHgCsr, MfgHgCsr]
CuGraphFgHgCsr = FgHgCsr
CuGraphMfgHgCsr = MfgHgCsr
CuGraphAnyHgCsr = AnyHgCsr
